#include "assert.h"

#include "stdio.h"
#include "string.h"
#include "math.h"
#include "stdlib.h"

#include "SDL.h"
#include "SDL_rotozoom.h"
#include "SDL_ttf.h"

#include "collision.h"
#include "BList.h"
#include "variables.h"
#include "mog2.h"
#include "tiles.h"

#include "auxiliar.h"
#include "reflection_effect.h"
#include "refraction_effect.h"

#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif
 

#define SINCOS_TABLE_SIZE	4096

extern int surface_flags;

TILE::TILE(void)
{
	r.x=0;
	r.y=0;
	r.h=0;
	r.w=0;
	orig=0;
	mask_visualization=0;
	mask_collision=0;
	base_collision_data=0;
	transform_collision_data=0;
	mask_type=0;
	collision_mask_type=0;
	tcd_angle=0;
	tcd_scale=0;
} /* TILE::TILE */ 


TILE::TILE(int x,int y,int dx,int dy,SDL_Surface *o,char *fn)
{

	r.x=x;
	r.y=y;
	r.h=dy;
	r.w=dx;

	orig=SDL_CreateRGBSurface(surface_flags,dx,dy,32,0,0,0,AMASK);
	SDL_SetAlpha(orig,0,SDL_ALPHA_OPAQUE);
	SDL_BlitSurface(o,&r,orig,0);
	orig_name=fn;
	mask_visualization=0;
	mask_type=0;

	mask_collision=SDL_CreateRGBSurface(0,dx,dy,32,0,0,0,AMASK);
	SDL_FillRect(mask_collision,0,0);
	surface_automatic_alpha(mask_collision);
	base_collision_data=MOG2_make_cmap(mask_collision);
	transform_collision_data=0;
	collision_mask_type=0;

	tcd_angle=0;
	tcd_scale=0;
} /* TILE::TILE */ 


TILE::~TILE(void)
{
	free();
} /* TILE::TILE */ 


bool TILE::save(FILE *fp)
{
	fprintf(fp,"%s\n",orig_name);
	fprintf(fp,"%i %i %i %i\n",r.x,r.y,r.w,r.h);
	fprintf(fp,"%i %i\n",mask_type,collision_mask_type);

	return true;
} /* TILE::save */ 


bool TILE::load(FILE *fp,MOG2 *game)
{
	List<TILE_SOURCE> l;
	TILE_SOURCE *t;
	char tmp[256];
	bool found=false;
	unsigned int i;
	int i1,i2,i3,i4;

	orig_name=0;
	orig=0;

	if (1!=fscanf(fp,"%s",tmp)) return false;
	for(i=0;i<strlen(tmp);i++) if (tmp[i]=='\\') tmp[i]='/';
	strlwr(tmp);

	if (6!=fscanf(fp,"%i %i %i %i %i %i",&i1,&i2,&i3,&i4,&mask_type,&collision_mask_type)) return false;

	r.x=i1;
	r.y=i2;
	r.w=i3;
	r.h=i4;

	l.Instance(game->tile_sources);
	l.Rewind();
	while(!found && l.Iterate(t)) {
		if (t->cmp(tmp)) {
			orig_name=t->fname;
			orig=SDL_CreateRGBSurface(surface_flags,r.w,r.h,32,0,0,0,AMASK);
			SDL_SetAlpha(orig,0,SDL_ALPHA_OPAQUE);
			if (t->sfc!=0) SDL_BlitSurface(t->sfc,&r,orig,0);
			found=true;
		} /* if */ 
	} /* while */ 


	if (mask_type==0) {
		SDL_SetAlpha(orig,0,SDL_ALPHA_OPAQUE);
	} /* if */ 

	if (mask_type==1) {
		SDL_SetAlpha(orig,SDL_SRCALPHA,SDL_ALPHA_OPAQUE);
		surface_automatic_alpha(orig);
	} /* if */ 

	if (mask_type==2) {
		SDL_SetAlpha(orig,SDL_SRCALPHA,SDL_ALPHA_OPAQUE);
		if (t->sfc!=0) surface_mask_from_bitmap(orig,t->sfc,r.x+r.w,r.y);
	} /* if */ 

	if (collision_mask_type==0) {
		if (mask_collision==0) mask_collision=SDL_CreateRGBSurface(surface_flags,r.w,r.h,32,0,0,0,AMASK);
		SDL_FillRect(mask_collision,0,0x00000000);
		surface_automatic_alpha(mask_collision);
		if (base_collision_data!=0) MOG2_destroy_cmap(base_collision_data);
		base_collision_data=MOG2_make_cmap(mask_collision);
	} /* if */ 

	if (collision_mask_type==1) {
		if (mask_collision==0) mask_collision=SDL_CreateRGBSurface(surface_flags,r.w,r.h,32,0,0,0,AMASK);
		SDL_FillRect(mask_collision,0,0xffffffff);
		surface_automatic_alpha(mask_collision);
		if (base_collision_data!=0) MOG2_destroy_cmap(base_collision_data);
		base_collision_data=MOG2_make_cmap(mask_collision);		
	} /* if */ 
						
	if (collision_mask_type==2) {
		SDL_Rect o;

		o.x=r.x+(r.w*2);
		o.y=r.y;
		o.w=r.w;
		o.h=r.h;
		if (mask_collision==0) mask_collision=SDL_CreateRGBSurface(surface_flags,r.w,r.h,32,0,0,0,AMASK);
		if (t->sfc!=0) SDL_BlitSurface(t->sfc,&o,mask_collision,0);
		surface_automatic_alpha(mask_collision);
		if (base_collision_data!=0) MOG2_destroy_cmap(base_collision_data);
		base_collision_data=MOG2_make_cmap(mask_collision);
	} /* if */ 

	if (collision_mask_type==3) {
		SDL_Rect o;

		o.x=r.x+(r.w);
		o.y=r.y;
		o.w=r.w;
		o.h=r.h;
		if (mask_collision==0) mask_collision=SDL_CreateRGBSurface(surface_flags,r.w,r.h,32,0,0,0,AMASK);
		if (t->sfc!=0) SDL_BlitSurface(t->sfc,&o,mask_collision,0);
		surface_automatic_alpha(mask_collision);
		if (base_collision_data!=0) MOG2_destroy_cmap(base_collision_data);
		base_collision_data=MOG2_make_cmap(mask_collision);
	} /* if */ 

	assert(base_collision_data!=0);

	return true;
} /* TILE:load */ 


void TILE::draw(int x,int y,SDL_Surface *dest)
{
	SDL_Rect d;

	if (orig!=0) {
		d.x=x;
		d.y=y;
		d.w=r.w;
		d.h=r.h;
		SDL_BlitSurface(orig,0,dest,&d);
	} /* if */ 
} /* TILE::draw */ 


void TILE::draw_shaded(int x,int y,SDL_Surface *dest,int factor,int red,int green,int blue,int alpha)
{
	SDL_Rect d;

	if (orig!=0) {
		SDL_Surface *tmp;
		tmp=SDL_DisplayFormatAlpha(orig);
		surface_shader(tmp,float(factor)/100.0F,red,green,blue,alpha,0);
		d.x=x;
		d.y=y;
		d.w=r.w;
		d.h=r.h;
		SDL_BlitSurface(tmp,0,dest,&d);
		SDL_FreeSurface(tmp);
	} /* if */ 
} /* TILE::draw_shaded */ 


void TILE::draw_bicolor(int x,int y,SDL_Surface *dest,int factor,int r1,int g1,int b1,int a1,int r2,int g2,int b2,int a2)
{
	SDL_Rect d;

	if (orig!=0) {
		SDL_Surface *tmp;
		tmp=SDL_DisplayFormatAlpha(orig);
		surface_bicolor(tmp,float(factor)/100.0F,r1,g1,b1,a1,r2,g2,b2,a2,0);
		d.x=x;
		d.y=y;
		d.w=r.w;
		d.h=r.h;
		SDL_BlitSurface(tmp,0,dest,&d);
		SDL_FreeSurface(tmp);
	} /* if */ 
} /* TILE::draw_bicolor */ 



void TILE::draw_transform(int x,int y,SDL_Surface *dest,float scale,float angle)
{

	if (orig!=0) {
		SDL_Rect d;
		SDL_Surface *res=rotozoomSurface(orig, (angle*180.0F)/M_PI, scale, SMOOTHING_ON);

		d.x=x;
		d.y=y;
		d.w=res->w;
		d.h=res->h;
		SDL_BlitSurface(res,0,dest,&d);

		SDL_FreeSurface(res);
	} /* if */ 
 
} /* TILE::draw_transform */ 


void TILE::draw_shaded_transform(int x,int y,SDL_Surface *dest,float scale,float angle,int factor,int red,int green,int blue,int alpha)
{

	if (orig!=0) {
		SDL_Rect d;
		SDL_Surface *res=rotozoomSurface(orig, (angle*180.0F)/M_PI, scale, SMOOTHING_ON);
		
		surface_shader(res,float(factor)/100.0F,red,green,blue,alpha,0);

		d.x=x;
		d.y=y;
		d.w=res->w;
		d.h=res->h;
		SDL_BlitSurface(res,0,dest,&d);

		SDL_FreeSurface(res);
	} /* if */ 
 
} /* TILE::draw_shaded_transform */ 


void TILE::draw_bicolor_transform(int x,int y,SDL_Surface *dest,float scale,float angle,int factor,int r1,int g1,int b1,int a1,int r2,int g2,int b2,int a2)
{
	if (orig!=0) {
		SDL_Rect d;
		SDL_Surface *res=rotozoomSurface(orig, (angle*180.0F)/M_PI, scale, SMOOTHING_ON);

		surface_bicolor(res,float(factor)/100.0F,r1,g1,b1,a1,r2,g2,b2,a2,0);

		d.x=x;
		d.y=y;
		d.w=res->w;
		d.h=res->h;
		SDL_BlitSurface(res,0,dest,&d);

		SDL_FreeSurface(res);
	} /* if */ 
} /* TILE::draw_bicolor_transform */ 




void TILE::draw_reflection(int x,int y,SDL_Surface *dest,int reflection_kh,int reflection_kv,int reflection_oh,int reflection_ov,int reflection_wave_height,int center_x,int center_y,int factor)
{

	if (orig!=0) {
		SDL_Rect d;
		SDL_Surface *reflection=SDL_CreateRGBSurface(surface_flags,get_dx(),get_dy(),32,0,0,0,0);;

		/* We will assume that the y is > bigger than SCREEN_Y/2: */ 
		if (y>(dest->h/2)) {
			float kh=3250.0F;	/* 5.0F */ 
			float kv=6500.0F;	/* 10.0F */ 
			float oh=0.0F;
			float ov=0.0F;

			if (reflection_kh!=0) kh=float(reflection_kh)*6.5F;
			if (reflection_kv!=0) kv=float(reflection_kv)*6.5F;
			if (reflection_oh!=0) oh=float(reflection_oh)*6.5F;
			if (reflection_ov!=0) ov=float(reflection_ov)*6.5F;
			reflection_draw(dest,y,reflection,kh,kv,oh,ov,float(reflection_wave_height),center_x,center_y);
		} else {
			SDL_FillRect(reflection,0,0);
		} /* if */ 

		d.x=x;
		d.y=y;
		d.w=reflection->w;
		d.h=reflection->h;
		SDL_BlitSurface(orig,0,dest,&d);

		factor=(factor*255)/100;
		if (factor<0) factor=0;
		if (factor>255) factor=255;
		SDL_SetAlpha(reflection, SDL_SRCALPHA, factor);
		SDL_BlitSurface(reflection,0,dest,&d);
		SDL_FreeSurface(reflection);
	} /* if */ 

} /* TILE::draw_reflection */ 


void TILE::draw_reflection_transform(int x,int y,SDL_Surface *dest,float scale,float angle,int reflection_kh,int reflection_kv,int reflection_oh,int reflection_ov,int reflection_wave_height,int center_x,int center_y,int factor)
{
	if (orig!=0) {
		SDL_Rect d;
		SDL_Surface *reflection=SDL_CreateRGBSurface(surface_flags,get_dx(),get_dy(),32,0,0,0,0);;
		/* We will assume that the y is > bigger than SCREEN_Y/2: */ 
		if (y>(dest->h/2)) {
			float kh=3250.0F;	/* 5.0F */ 
			float kv=6500.0F;	/* 10.0F */ 
			float oh=0.0F;
			float ov=0.0F;

			if (reflection_kh!=0) kh=float(reflection_kh)*6.5F;
			if (reflection_kv!=0) kv=float(reflection_kv)*6.5F;
			if (reflection_oh!=0) oh=float(reflection_oh)*6.5F;
			if (reflection_ov!=0) ov=float(reflection_ov)*6.5F;
			reflection_draw(dest,y,reflection,kh,kv,oh,ov,float(reflection_wave_height),center_x,center_y);
		} else {
			SDL_FillRect(reflection,0,0);
		} /* if */ 
		SDL_Surface *res=rotozoomSurface(reflection, (angle*180.0F)/M_PI, scale, SMOOTHING_ON);

		d.x=x;
		d.y=y;
		d.w=res->w;
		d.h=res->h;
		SDL_BlitSurface(res,0,dest,&d);

		SDL_FreeSurface(res);
		SDL_FreeSurface(reflection);
	} /* if */ 

} /* TILE::draw_reflection_transform */ 



void TILE::draw_refraction(int x,int y,SDL_Surface *dest,int refraction_distance,int refraction_kh,int refraction_kv,int refraction_oh,int refraction_ov,int refraction_wave_height,int factor)
{

	if (orig!=0) {
		SDL_Rect d;
		float kh=16.0F;	/* 16.0F */ 
		float kv=8.0F;	/* 8.0F */ 
		float oh=0.0F;
		float ov=0.0F;
		float dist=0.0F;
		SDL_Surface *refraction=SDL_CreateRGBSurface(surface_flags,get_dx(),get_dy(),32,0,0,0,0);;

		if (refraction_kh!=0) kh=float(refraction_kh)/100;
		if (refraction_kv!=0) kv=float(refraction_kv)/100;
		if (refraction_oh!=0) oh=float(refraction_oh)/100;
		if (refraction_ov!=0) ov=float(refraction_ov)/100;
		if (refraction_distance!=0) dist=float(refraction_distance)/100;
		
		d.x=x;
		d.y=y;
		d.w=refraction->w;
		d.h=refraction->h;
		refraction_draw(dest,d,refraction,dist,kh,kv,oh,ov,float(refraction_wave_height));
//		SDL_BlitSurface(dest,&d,refraction,0);

		SDL_BlitSurface(orig,0,dest,&d);

		factor=(factor*255)/100;
		if (factor<0) factor=0;
		if (factor>255) factor=255;
		SDL_SetAlpha(refraction, SDL_SRCALPHA, factor);
		SDL_BlitSurface(refraction,0,dest,&d);
		SDL_FreeSurface(refraction);
	} /* if */ 

} /* TILE::draw_refraction */ 


void TILE::draw_refraction_transform(int x,int y,SDL_Surface *dest,float scale,float angle,int refraction_distance,int refraction_kh,int refraction_kv,int refraction_oh,int refraction_ov,int refraction_wave_height,int factor)
{
	if (orig!=0) {
		SDL_Rect d;
		float kh=16.0F;	/* 16.0F */ 
		float kv=8.0F;	/* 8.0F */ 
		float oh=0.0F;
		float ov=0.0F;
		float dist=0.0F;
		SDL_Surface *refraction=SDL_CreateRGBSurface(surface_flags,get_dx(),get_dy(),32,0,0,0,0);;

		if (refraction_kh!=0) kh=float(refraction_kh)/100;
		if (refraction_kv!=0) kv=float(refraction_kv)/100;
		if (refraction_oh!=0) oh=float(refraction_oh)/100;
		if (refraction_ov!=0) ov=float(refraction_ov)/100;
		if (refraction_distance!=0) dist=float(refraction_distance)/100;

		d.x=x;
		d.y=y;
		d.w=get_dx();
		d.h=get_dy();
		refraction_draw(dest,d,refraction,dist,kh,kv,oh,ov,float(refraction_wave_height));
		SDL_Surface *res=rotozoomSurface(refraction, (angle*180.0F)/M_PI, scale, SMOOTHING_ON);

		d.x=x;
		d.y=y;
		d.w=res->w;
		d.h=res->h;
		SDL_BlitSurface(res,0,dest,&d);

		SDL_FreeSurface(res);
		SDL_FreeSurface(refraction);
	} /* if */ 

} /* TILE::draw_refraction_transform */ 


void TILE::draw_collision_mask(int x,int y,SDL_Surface *dest)
{
	SDL_Rect d;

	if (orig!=0) {
		if (mask_collision!=0) {
			d.x=x;
			d.y=y;
			d.w=r.w;
			d.h=r.h;
			SDL_BlitSurface(mask_collision,0,dest,&d);
		} /* if */ 
	} /* if */ 
} /* TILE::draw_collision_mask */ 


void TILE::draw_collision_mask_transform(int x,int y,SDL_Surface *dest,float scale,float angle)
{
	SDL_Rect d;

	if (orig!=0) {
		if (mask_collision!=0) {
			SDL_Surface *res=rotozoomSurface(mask_collision, (angle*180.0F)/M_PI, scale, SMOOTHING_OFF);

			d.x=x;
			d.y=y;
			d.w=res->w;
			d.h=res->h;
			SDL_BlitSurface(res,0,dest,&d);

			SDL_FreeSurface(res);
		} /* if */ 
	} /* if */ 
} /* TILE::draw_collision_mask_transform */ 



void TILE::draw_mask(int x,int y,SDL_Surface *dest)
{
	SDL_Rect d;

	if (orig!=0) {
		if (mask_visualization==0) {
			int i,j;

			mask_visualization=SDL_CreateRGBSurface(surface_flags,r.w,r.h,32,0,0,0,0);
			for(i=0;i<r.w;i++) {
				for(j=0;j<r.h;j++) {
					Uint32 color;
                    Uint8 r,g,b,a;
                    
                    color=getpixel(orig,i,j);
                    SDL_GetRGBA(color,orig->format,&r,&g,&b,&a);

                    color=SDL_MapRGBA(mask_visualization->format,a,a,a,0);
                    putpixel(mask_visualization,i,j,color);		
				} /* for */ 
			} /* for */ 
		} /* if */ 

		d.x=x;
		d.y=y;
		d.w=r.w;
		d.h=r.h;
		SDL_BlitSurface(mask_visualization,0,dest,&d);
	} /* if */ 
} /* TILE::draw_mask */ 


void TILE::clear(void)
{
	orig=0;
	mask_visualization=0;
	mask_collision=0;
	base_collision_data=0;
	transform_collision_data=0;
	orig_name=0;
} /* TILE::clear */ 


void TILE::free(void)
{
	if (orig!=0) SDL_FreeSurface(orig);
	orig=0;
	if (mask_visualization!=0) SDL_FreeSurface(mask_visualization);
	mask_visualization=0;

	if (mask_collision!=0) SDL_FreeSurface(mask_collision);
	mask_collision=0;

	if (base_collision_data!=0) MOG2_destroy_cmap(base_collision_data);
	base_collision_data=0;

	if (transform_collision_data!=0) MOG2_destroy_cmap(transform_collision_data);
	transform_collision_data=0;
} /* TILE::free */ 


void TILE::instance(TILE *t)
{
	r=t->r;
	orig=t->orig;
	mask_visualization=t->mask_visualization;
	mask_collision=t->mask_collision;
	base_collision_data=t->base_collision_data;
	transform_collision_data=t->transform_collision_data;

	orig_name=t->orig_name;
	mask_type=t->mask_type;
	collision_mask_type=t->collision_mask_type;

} /* TILE::instace */ 


MOG2CollisionData *TILE::get_collision_data(float scale,float angle)
{
	if (scale==1.0 && angle==0) {
		return base_collision_data;
	} else {
		if (transform_collision_data==0 ||
			scale!=tcd_scale ||
			angle!=tcd_angle) {
			SDL_Surface *res=rotozoomSurface(mask_collision, (angle*180.0F)/M_PI, scale, SMOOTHING_OFF);

			if (transform_collision_data!=0) MOG2_destroy_cmap(transform_collision_data);
			transform_collision_data=0;

			transform_collision_data=MOG2_make_cmap(res);
			tcd_scale=scale;
			tcd_angle=angle;
			SDL_FreeSurface(res);
		} /* if */ 

		return transform_collision_data;
	} /* if */ 
} /* get_collision_data */ 



